import pytest
import mock
import numpy as np
import gilp.simplex as sm


class TestLP:

    def test_init_exceptions(self):
        with pytest.raises(ValueError, match='.* b should have shape .*'):
            A = np.array([[1,0],[0,1]])
            b = np.array([[1],[2],[3]])
            c = np.array([[1],[2]])
            sm.LP(A,b,c)
        with pytest.raises(ValueError, match='.*not nonnegative.*'):
            A = np.array([[1,0],[0,1]])
            b = np.array([[1],[-2]])
            c = np.array([[1],[2]])
            sm.LP(A,b,c)
        with pytest.raises(ValueError, match='.* c should have shape .*'):
            A = np.array([[1,0],[0,1]])
            b = np.array([[1],[2]])
            c = np.array([[1],[2],[3]])
            sm.LP(A,b,c)

    @pytest.mark.parametrize("n,m,A,b,c,A_I,c_0",[
        (2,2,
         np.array([[1,2],[3,0]]),
         np.array([[3],[4]]),
         np.array([[1],[2]]),
         np.array([[1,2,1,0],[3,0,0,1]]),
         np.array([[1],[2],[0],[0]])),
        (3,2,
         np.array([[1,2,3],[3,0,1]]),
         np.array([[3],[4]]),
         np.array([[1],[2],[3]]),
         np.array([[1,2,3,1,0],[3,0,1,0,1]]),
         np.array([[1],[2],[3],[0],[0]]))])
    def test_init(self,n,m,A,b,c,A_I,c_0):
        lp = sm.LP(A,b,c)
        actual = lp.get_inequality_form()
        assert n == actual[0]
        assert m == actual[1]
        assert (A == actual[2]).all()
        assert (b == actual[3]).all()
        assert (c == actual[4]).all()
        actual = lp.get_equality_form()
        assert n == actual[0]
        assert m == actual[1]
        assert (A_I == actual[2]).all()
        assert (b == actual[3]).all()
        assert (c_0 == actual[4]).all()

    def test_get_bfs(self, degenerate_lp):
        lp = degenerate_lp
        bfs = np.array([[2],[4],[0],[0],[4],[1],[0]])
        assert (bfs == lp.get_basic_feasible_sol([0,1,4,5,6])).all()
        assert (bfs == lp.get_basic_feasible_sol([0,1,2,4,5])).all()
        assert (bfs == lp.get_basic_feasible_sol([0,1,3,4,5])).all()
        with pytest.raises(sm.InvalidBasis):
            lp.get_basic_feasible_sol([1,2,3,4])
        with pytest.raises(sm.InvalidBasis):
            lp.get_basic_feasible_sol([0,1,2,4,5,6])
        with pytest.raises(sm.InfeasibleBasicSolution):
            lp.get_basic_feasible_sol([0,1,2,3,5])

    def test_get_all_bfs(self):
        A = np.array([[1,1],[-2,1]])
        b = np.array([[6],[0]])
        c = np.array([[1],[1]])
        lp = sm.LP(A,b,c)
        bfs = [np.array([[2],[4],[0],[0]]),
               np.array([[0],[0],[6],[0]]),
               np.array([[6],[0],[0],[12]]),
               np.array([[0],[0],[6],[0]]),
               np.array([[0],[0],[6],[0]])]
        bases = [[0,1],[0,2],[0,3],
                 [1,2],[2,3]]
        values = [6,0,6,0,0]
        actual = lp.get_basic_feasible_solns()

        assert all(np.allclose(x,y,atol=1e-7) for x,y in zip(bfs, actual[0]))
        assert bases == actual[1]
        assert values == actual[2]

    def test_tableau(self, degenerate_lp):
        T = np.array([[1,0,0,-1,-1,0,0,0,10],
                      [0,1,0,1,-1,0,0,0,2],
                      [0,0,1,0,1,0,0,0,4],
                      [0,0,0,-1,2,1,0,0,4],
                      [0,0,0,-1,1,0,1,0,1],
                      [0,0,0,2,-3,0,0,1,0]])
        assert (T == degenerate_lp.get_tableau([0,1,4,5,6])).all()
        with pytest.raises(sm.InvalidBasis):
            degenerate_lp.get_tableau([1,2,3])


class TestSimplexIteration:

    def test_bad_inputs(self, klee_minty_3d_lp):
        with pytest.raises(ValueError,match='Invalid pivot rule.*'):
            sm.simplex_iteration(lp=klee_minty_3d_lp,
                                 x=np.array([[5],[5],[0],[0],[0],[65]]),
                                 B=[0,1,5],
                                 pivot_rule='invalid')
        with pytest.raises(ValueError,match='x should have shape.*'):
            sm.simplex_iteration(lp=klee_minty_3d_lp,
                                 x=np.array([[5],[5],[0],[0],[0]]),
                                 B=[0,1,5],
                                 pivot_rule='bland')
        with pytest.raises(ValueError,match='.*different basic feasible.*'):
            sm.simplex_iteration(lp=klee_minty_3d_lp,
                                 x=np.array([[5],[5],[0],[0],[0],[65]]),
                                 B=[0,1,2],
                                 pivot_rule='bland')

    def test_bland(self, klee_minty_3d_lp):
        actual = sm.simplex_iteration(lp=klee_minty_3d_lp,
                                      x=np.array([[5],[5],[0],[0],[0],[65]]),
                                      B=[0,1,5],
                                      pivot_rule='bland')
        assert (np.array([[5],[5],[65],[0],[0],[0]]) == actual[0]).all()
        actual[1].sort()
        assert [0,1,2] == actual[1]
        assert 95 == actual[2]
        assert not actual[3]

    def test_min_index(self, klee_minty_3d_lp):
        actual = sm.simplex_iteration(lp=klee_minty_3d_lp,
                                      x=np.array([[5],[5],[0],[0],[0],[65]]),
                                      B=[0,1,5],
                                      pivot_rule='min_index')
        assert (np.array([[5],[5],[65],[0],[0],[0]]) == actual[0]).all()
        actual[1].sort()
        assert [0,1,2] == actual[1]
        assert 95 == actual[2]
        assert not actual[3]

    def test_dantzig(self, klee_minty_3d_lp):
        actual = sm.simplex_iteration(lp=klee_minty_3d_lp,
                                      x=np.array([[0],[0],[0],[5],[25],[125]]),
                                      B=[3,4,5],
                                      pivot_rule='dantzig')
        assert (np.array([[5],[0],[0],[0],[5],[85]]) == actual[0]).all()
        actual[1].sort()
        assert [0,4,5] == actual[1]
        assert 20 == actual[2]
        assert not actual[3]

    def test_max_reduced_cost(self, klee_minty_3d_lp):
        actual = sm.simplex_iteration(lp=klee_minty_3d_lp,
                                      x=np.array([[0],[0],[0],[5],[25],[125]]),
                                      B=[3,4,5],
                                      pivot_rule='max_reduced_cost')
        assert (np.array([[5],[0],[0],[0],[5],[85]]) == actual[0]).all()
        actual[1].sort()
        assert [0,4,5] == actual[1]
        assert 20 == actual[2]
        assert not actual[3]

    def test_greatest_ascent1(self, klee_minty_3d_lp):
        actual = sm.simplex_iteration(lp=klee_minty_3d_lp,
                                      x=np.array([[0],[0],[0],[5],[25],[125]]),
                                      B=[3,4,5],
                                      pivot_rule='greatest_ascent')
        assert (np.array([[0],[0],[125],[5],[25],[0]]) == actual[0]).all()
        actual[1].sort()
        assert [2,3,4] == actual[1]
        assert 125 == actual[2]
        assert not actual[3]

    def test_greatest_ascent2(self, klee_minty_3d_lp):
        actual = sm.simplex_iteration(lp=klee_minty_3d_lp,
                                      x=np.array([[0],[0],[125],[5],[25],[0]]),
                                      B=[2,3,4],
                                      pivot_rule='greatest_ascent')
        assert (np.array([[0],[0],[125],[5],[25],[0]]) == actual[0]).all()
        actual[1].sort()
        assert [2,3,4] == actual[1]
        assert 125 == actual[2]
        assert actual[3]

    def test_manual_select(self, klee_minty_3d_lp):
        with mock.patch('builtins.input', return_value="1"):
            actual = sm.simplex_iteration(lp=klee_minty_3d_lp,
                                          x=np.array([[0],[0],[0],
                                                      [5],[25],[125]]),
                                          B=[3,4,5],
                                          pivot_rule='manual_select')
            assert (np.array([[0],[25],[0],[5],[0],[25]]) == actual[0]).all()
            actual[1].sort()
            assert [1,3,5] == actual[1]
            assert 50 == actual[2]
            assert not actual[3]


class TestSimplex():

    def test_bad_inputs(self, klee_minty_3d_lp, unbounded_lp):
        with pytest.raises(ValueError,match='Invalid pivot rule.*'):
            sm.simplex(lp=klee_minty_3d_lp,
                       pivot_rule='invalid')
        with pytest.raises(ValueError,match='.*should have shape.*'):
            sm.simplex(lp=klee_minty_3d_lp,
                       initial_solution=np.array([[5],[5],[0],[0]]))
        with pytest.raises(ValueError,match='Iteration limit*'):
            sm.simplex(lp=klee_minty_3d_lp,
                       iteration_limit=-1)
        with pytest.raises(sm.UnboundedLinearProgram):
            # Make sure the initial solution is ignored and no error is raised
            sm.simplex(unbounded_lp,initial_solution=np.array([[2],[2]]))
        with pytest.raises(sm.UnboundedLinearProgram):
            sm.simplex(unbounded_lp,'greatest_ascent')

    def test_simplex(self, klee_minty_3d_lp):
        actual = sm.simplex(klee_minty_3d_lp,pivot_rule='dantzig')
        bfs = [np.array([[0],[0],[0],[5],[25],[125]]),
               np.array([[5],[0],[0],[0],[5],[85]]),
               np.array([[5],[5],[0],[0],[0],[65]]),
               np.array([[0],[25],[0],[5],[0],[25]]),
               np.array([[0],[25],[25],[5],[0],[0]]),
               np.array([[5],[5],[65],[0],[0],[0]]),
               np.array([[5],[0],[85],[0],[5],[0]]),
               np.array([[0],[0],[125],[5],[25],[0]])]
        bases = [[3,4,5], [0,4,5], [0,1,5],
                 [1,3,5], [1,2,3], [0,1,2],
                 [0,2,4], [2,3,4]]
        assert all(np.allclose(x,y,atol=1e-7) for x,y in zip(bfs, actual[0]))
        for basis in actual[1]:
            basis.sort()
        assert bases == actual[1]
        assert 125 == actual[2]
        assert actual[3]

    def test_initial_solution(self, klee_minty_3d_lp):
        actual = sm.simplex(klee_minty_3d_lp,
                            initial_solution=np.array([[5],[5],[65]]),
                            pivot_rule='dantzig')
        bfs = [np.array([[5],[5],[65],[0],[0],[0]]),
               np.array([[5],[0],[85],[0],[5],[0]]),
               np.array([[0],[0],[125],[5],[25],[0]])]
        bases = [[0,1,2], [0,2,4], [2,3,4]]
        assert all(np.allclose(x,y,atol=1e-7) for x,y in zip(bfs, actual[0]))
        for basis in actual[1]:
            basis.sort()
        assert bases == actual[1]
        assert 125 == actual[2]
        assert actual[3]

    def test_degenerate_init_sol(self):
        A = np.array([[1,1],[0,1],[1,-1],[1,0],[-2,1]])
        b = np.array([[6],[4],[2],[3],[0]])
        c = np.array([[1],[0]])
        lp = sm.LP(A,b,c)
        sm.simplex(lp,initial_solution=np.array([[2],[4]]))

    def test_iteration_limit(self, klee_minty_3d_lp):
        actual = sm.simplex(klee_minty_3d_lp,
                            pivot_rule='dantzig',
                            iteration_limit=3)
        bfs = [np.array([[0],[0],[0],[5],[25],[125]]),
               np.array([[5],[0],[0],[0],[5],[85]]),
               np.array([[5],[5],[0],[0],[0],[65]]),
               np.array([[0],[25],[0],[5],[0],[25]])]
        bases = [[3,4,5], [0,4,5], [0,1,5], [1,3,5]]
        assert all(np.allclose(x,y,atol=1e-7) for x,y in zip(bfs, actual[0]))
        for basis in actual[1]:
            basis.sort()
        assert bases == actual[1]
        assert 50 == actual[2]
        assert not actual[3]


@pytest.mark.parametrize("A,t",[
    (np.array([[1,0],[0,1]]), True),
    (np.array([[0,1],[1,0]]), True),
    (np.array([[1,0,0],[0,1,0]]), False),
    (np.array([[2,0,0],[0,0,3],[0,1,0]]), True)])
def test_invertible(A,t):
    assert sm.invertible(A) == t



import numpy as np
import itertools
from scipy.linalg import solve
from typing import List, Tuple

"""Provides an implementation of the revised simplex method.

Classes:
    UnboundedLinearProgram: Exception indicating the unboundedness of an LP.
    InvalidBasis: Exception indicating a list of indices does not form a basis.
    InfeasibleBasicSolution: Exception indicating infeasible bfs.
    LP: Maintains the coefficents and size of a linear program (LP).

Functions:
    invertible: Return true if the matrix A is invertible.
    simplex_iter: Execute a single iteration of the revised simplex method.
    simplex: Run the revised simplex method.
"""


class UnboundedLinearProgram(Exception):
    """Raised when an LP is found to be unbounded during an execution of the
    revised simplex method"""
    pass


class InvalidBasis(Exception):
    """Raised when a list of indices does not form a valid basis and prevents
    further correct execution of the function."""
    pass


class InfeasibleBasicSolution(Exception):
    """Raised when a list of indices forms a valid basis but the corresponding
    basic solution is infeasible."""
    pass


class LP:
    """Maintains the coefficents and size of a linear program (LP).

    The LP class maintains the coefficents of a linear program in both standard
    inequality and equality form. A is an m*n matrix describing the linear
    combination of variables making up the LHS of each constraint. b is a
    nonnegative vector of length m making up the RHS of each constraint.
    Lastly, c is a vector of length n describing the objective function to be
    maximized. Both the n decision variables and m slack variables must be
    nonnegative. Under these assumptions, the LP must be feasible.

    ::

        inequality        equality
        max c^Tx          max c^Tx
        s.t Ax <= b       s.t Ax + Is == b
            x >= 0               x,s >= 0

    Attributes:
        n (int): number of decision variables (excluding slack variables).
        m (int): number of constraints (excluding nonnegativity constraints).
        A (np.ndarray): An m*n matrix of coefficients.
        A_I (np.ndarray): An m*(n+m) matrix of coefficients: [A I].
        b (np.ndarray): A nonnegative vector of coefficients of length m.
        c (np.ndarray): A vector of coefficients of length n.
        c_0 (np.ndarray): A vector of coefficients of length n+m: [c^T 0^T]^T.
    """

    def __init__(self, A: np.ndarray, b: np.ndarray, c: np.ndarray):
        """Initializes an LP.

        Creates an instance of LP using the given coefficents. Note: the given
        coefficents must correspond to an LP in standard INEQUALITY form::

            max c^Tx
            s.t Ax <= b
                x >= 0

        Args:
            A (np.ndarray): An m*n matrix of coefficients.
            b (np.ndarray): A nonnegative vector of coefficients of length m.
            c (np.ndarray): A vector of coefficients of length n.

        Raises:
            ValueError: A has shape (m,n). b should have shape (m,1) but was ().
            ValueError: b is not nonnegative. Was [].
            ValueError: A has shape (m,n). c should have shape (n,1) but was ().
        """
        self.m = len(A)
        self.n = len(A[0])
        if not b.shape == (self.m, 1):
            raise ValueError('A has shape ' + str(A.shape)
                             + '. b should have shape (' + str(self.m) + ',1) '
                             + 'but was ' + str(b.shape) + '.')
        if not all(b >= np.zeros((self.m, 1))):
            raise ValueError('b is not nonnegative. Was \n'+str(b))
        if not c.shape == (self.n, 1):
            raise ValueError('A has shape '+str(A.shape)
                             + '. c should have shape (' + str(self.n)+',1) '
                             + 'but was '+str(c.shape) + '.')
        self.A = A
        self.A_I = np.hstack((self.A, np.identity(self.m)))
        self.b = b
        self.c = c
        self.c_0 = np.vstack((self.c, np.zeros((self.m, 1))))

    def get_inequality_form(self):
        """Returns n,m,A,b,c describing this LP in standard inequality form."""
        return self.n, self.m, self.A, self.b, self.c

    def get_equality_form(self):
        """Returns n,m,A_I,b,c_0 describing this LP in standard equality form"""
        return self.n, self.m, self.A_I, self.b, self.c_0

    def get_basic_feasible_sol(self, B: List[int]) -> np.ndarray:
        """Return the basic feasible solution corresponding to this basis.

        By definition, B is a basis iff A_B is invertible (where A is the
        matrix of coefficents in standard equality form). The corresponding
        basic solution x satisfies A_Bx = b. By definition, x is a basic
        feasible solution iff x satisfies both A_Bx = b and x > 0.

        Args:
            B (List[int]): A list of indices in {0..n+m-1} forming a basis.

        Returns:
            np.ndarray: Basic feasible solution corresponding to the basis B.

        Raises:
            InvalidBasis: B
            InfeasibleBasicSolution: x_B
        """
        n,m,A,b,c = self.get_equality_form()
        B.sort()
        if B[-1] < n+m and invertible(A[:,B]):
            x_B = np.zeros((n+m, 1))
            x_B[B,:] = np.round(solve(A[:,B], b), 7)
            if all(x_B >= np.zeros((n+m, 1))):
                return x_B
            else:
                raise InfeasibleBasicSolution(x_B)
        else:
            raise InvalidBasis(B)

    def get_basic_feasible_solns(self) -> Tuple[List[np.ndarray],
                                                List[List[int]],
                                                List[float]]:
        """Return all basic feasible solutions, their basis, and objective value.

        Returns:
            Tuple:

            - List[np.ndarray]: The list of basic feasible solutions for this LP.
            - List[List[int]]: The corresponding list of bases.
            - List[float]: The corresponding list of objective values.
        """
        n, m, A, b, c = self.get_equality_form()
        bfs, bases, values = [], [], []
        for B in itertools.combinations(range(n+m), m):
            try:
                x_B = self.get_basic_feasible_sol(list(B))
                bfs.append(x_B)
                bases.append(list(B))
                values.append(float(np.round(np.dot(c.transpose(), x_B), 7)))
            except (InvalidBasis, InfeasibleBasicSolution):
                pass
        return (bfs, bases, values)

    def get_tableau(self, B: List[int]) -> np.ndarray:
        """Return the tableau corresponding to the basis B for this LP.

        The returned tableau has the following form::

            z - (c_N^T - y^TA_N)x_N = y^Tb  where   y^T = c_B^TA_B^(-1)
            x_B + A_B^(-1)A_Nx_N = x_B^*    where   x_B^* = A_B^(-1)b

        Args:
            B (List[int]): A valid basis for this LP

        Returns:
            np.ndarray: A numpy array representing the tableau

        Raises:
            InvalidBasis: Invalid basis. A_B is not invertible.
        """
        n,m,A,b,c = self.get_equality_form()
        if not invertible(A[:,B]):
            raise InvalidBasis('Invalid basis. A_B is not invertible.')

        N = list(set(range(n+m)) - set(B))
        B.sort()
        N.sort()
        A_B_inv = np.linalg.inv(A[:,B])
        yT = np.dot(c[B,:].transpose(), A_B_inv)

        T = np.zeros((m+1, n+m+2))
        T[0,0] = 1
        T[0,1:n+m+1][N] = c[N,:].transpose() - np.dot(yT, A[:,N])
        T[0,n+m+1] = np.dot(yT,b)
        T[1:,1:n+m+1][:,N] = np.dot(A_B_inv, A[:,N])
        T[1:,1:n+m+1][:,B] = np.identity(len(B))
        T[1:,n+m+1] = np.dot(A_B_inv, b)[:,0]
        return np.round(T,7)


def invertible(A:np.ndarray) -> bool:
    """Return true if the matrix A is invertible.

    By definition, a matrix A is invertible iff n = m and A has rank n

    Args:
        A (np.ndarray): An m*n matrix

    Returns:
        bool: True if the matrix A is invertible. False otherwise.
    """
    return len(A) == len(A[0]) and np.linalg.matrix_rank(A) == len(A)


def simplex_iteration(lp: LP,
                      x: np.ndarray,
                      B: List[int],
                      pivot_rule: str = 'bland'
                      ) -> Tuple[np.ndarray, List[int], float, bool]:
    """Execute a single iteration of the revised simplex method.

    Let x be the initial basic feasible solution with corresponding basis B. Do
    one iteration of the revised simplex method using the given pivot rule.
    Implemented pivot rules include:

    Entering variable:

        - 'bland' or 'min_index': minimum index
        - 'dantzig' or 'max_reduced_cost': most positive reduced cost
        - 'greatest_ascent': most positive (minimum ratio) x (reduced cost)
        - 'manual_select': user selects among possible entering indices

    Leaving variable:

        - (All): minimum (positive) ratio (minimum index to tie break)

    Args:
        lp (LP): LP on which the simplex iteration is being done.
        x (np.ndarray): Initial basic feasible solution.
        B (List(int)): Basis corresponding to basic feasible solution x.
        pivot_rule (str): Pivot rule to be used. 'bland' by default.

    Returns:
        Tuple:

        - np.ndarray: New basic feasible solution.
        - List[int]: Basis corresponding to the new basic feasible solution.
        - float: Objective value of the new basic feasible solution.
        - bool: An idication of optimality. True if optimal. False otherwise.

    Raises:
        ValueError: Invalid pivot rule. Select from (list).
        ValueError: x should have shape (n+m,1) but was ().

    """
    pivot_rules = ['bland','min_index','dantzig','max_reduced_cost',
                   'greatest_ascent','manual_select']
    if pivot_rule not in pivot_rules:
        raise ValueError('Invalid pivot rule. Select from ' + str(pivot_rules))
    n,m,A,b,c = lp.get_equality_form()
    if not x.shape == (n+m, 1):
        raise ValueError('x should have shape (' + str(n+m) + ',1) '
                         + 'but was ' + str(x.shape))
    if not np.allclose(x, lp.get_basic_feasible_sol(B), atol=1e-07):
        raise ValueError('The basis ' + str(B) + ' corresponds to a different '
                         + 'basic feasible solution.')

    N = list(set(range(n+m)) - set(B))
    y = solve(A[:,B].transpose(), c[B,:])
    red_costs = c - np.dot(y.transpose(),A).transpose()
    entering = {k: red_costs[k] for k in N if red_costs[k] > 0}
    if len(entering) == 0:
        current_value = float(np.round(np.dot(c.transpose(), x), 7))
        return x,B,current_value,True
    else:
        if pivot_rule == 'greatest_ascent':
            eligible = {}
            for k in entering:
                d = np.zeros((1,n+m))
                d[:,B] = solve(A[:,B], A[:,k])
                ratios = {i: x[i]/d[0][i] for i in B if d[0][i] > 0}
                if len(ratios) == 0:
                    raise UnboundedLinearProgram('This LP is unbounded')
                t = min(ratios.values())
                r_pos = [r for r in ratios if ratios[r] == t]
                r = min(r_pos)
                t = ratios[r]
                eligible[(t*red_costs[k])[0]] = [k,r,t,d]
            k,r,t,d = eligible[max(eligible.keys())]
        else:
            user_input = None
            if pivot_rule == 'manual_select':
                user_input = int(input('Pick one of ' + str(entering.keys())))
            k = {'bland': min(entering.keys()),
                 'min_index': min(entering.keys()),
                 'dantzig': max(entering, key=entering.get),
                 'max_reduced_cost': max(entering, key=entering.get),
                 'manual_select': user_input}[pivot_rule]
            d = np.zeros((1,n+m))
            d[:,B] = solve(A[:,B], A[:,k])
            ratios = {i: x[i]/d[0][i] for i in B if d[0][i] > 0}
            if len(ratios) == 0:
                raise UnboundedLinearProgram('This LP is unbounded')
            t = min(ratios.values())
            r_pos = [r for r in ratios if ratios[r] == t]
            r = min(r_pos)
            t = ratios[r]
        # update
        x[k] = t
        x[B,:] = x[B,:] - t*(d[:,B].transpose())
        B.append(k)
        B.remove(r)
        N.append(r)
        N.remove(k)
        current_value = float(np.round(np.dot(c.transpose(), x), 7))
        return np.round(x,7),B,current_value,False


def simplex(lp: LP,
            pivot_rule: str = 'bland',
            initial_solution: np.ndarray = None,
            iteration_limit: int = None
            ) -> Tuple[List[np.ndarray], List[List[int]], float, bool]:
    """Execute the revised simplex method on the given LP.

    Execute the revised simplex method on the given LP using the specified
    pivot rule. If a valid initial basic feasible solution is given, use it as
    the initial bfs. Otherwise, ignore it. If an iteration limit is given,
    terminate if the specified limit is reached. Output the current solution
    and indicate the solution may not be optimal.

    PIVOT RULES

    Entering variable:

        - 'bland' or 'min_index': minimum index
        - 'dantzig' or 'max_reduced_cost': most positive reduced cost
        - 'greatest_ascent': most positive (minimum ratio) x (reduced cost)
        - 'manual_select': user selects among possible entering indices

    Leaving variable:

        - (All): minimum (positive) ratio (minimum index to tie break)

    Args:
        lp (LP): LP on which to run simplex
        pivot_rule (str): Pivot rule to be used. 'bland' by default.
        initial_solution (np.ndarray): Initial bfs. None by default.
        iteration_limit (int): Simplex iteration limit. None by default.

    Return:
        Tuple:

        - List[np.ndarray]: Basic feasible solutions at each simplex iteration.
        - List[List[int]]: Corresponding bases at each simplex iteration.
        - float: The current objective value.
        - bool: True if the current objective value is known to be optimal.

    Raises:
        ValueError: Invalid pivot rule. Select from (list).
        ValueError: Iteration limit must be strictly positive.
        ValueError: initial_solution should have shape (n,1) but was ().
    """
    pivot_rules = ['bland','min_index','dantzig','max_reduced_cost',
                   'greatest_ascent','manual_select']
    if pivot_rule not in pivot_rules:
        raise ValueError('Invalid pivot rule. Select from ' + str(pivot_rules))
    if iteration_limit is not None and iteration_limit <= 0:
        raise ValueError('Iteration limit must be strictly positive.')

    n,m,A,b,c = lp.get_equality_form()

    B = list(range(n,n+m))
    x = np.zeros((n+m,1))
    x[B,:] = b

    if initial_solution is not None:
        if not initial_solution.shape == (n, 1):
            raise ValueError('initial_solution should have shape (' + str(n)
                             + ',1) but was ' + str(initial_solution.shape))
        x_B = np.zeros((n+m,1))
        x_B[:n] = initial_solution
        x_B[n:] = b-np.dot(lp.A,initial_solution)
        x_B = np.round(x_B,7)
        if (all(np.round(np.dot(A,x_B)) == b) and
                all(x_B >= np.zeros((n+m,1))) and
                len(np.nonzero(x_B)[0]) <= m):
            x = x_B
            B = list(np.nonzero(x_B)[0])
            N = list(set(range(n+m)) - set(B))
            while len(B) < m:  # if initial solution is degenerate
                B.append(N.pop())
        else:
            print('Initial solution ignored.')

    path = [np.copy(x)]
    bases = [list.copy(B)]
    current_value = float(np.round(np.dot(c.transpose(), x), 7))
    optimal = False

    if iteration_limit is not None:
        lim = iteration_limit
    while(not optimal):
        x,B,value,opt = simplex_iteration(lp,x,B,pivot_rule)
        current_value = value
        if opt:
            optimal = True
        else:
            path.append(np.copy(x))
            bases.append(list.copy(B))
        if iteration_limit is not None:
            lim = lim - 1
        if iteration_limit is not None and lim == 0:
            break

    return path, bases, current_value, optimal



import numpy as np
import itertools
from scipy.spatial import ConvexHull
import plotly.graph_objects as plt
from .simplex import LP, InvalidBasis, InfeasibleBasicSolution
from typing import List, Dict, Union

"""Provides a higher level interface with plotly. Includes default styles.

Functions:
    format: Return a properly formated string for a number at some precision.
    linear_string: Return the string representation of a linear combination.
    equation_string: Return the string representation of an equation.
    label: Return a styled string representation of the given dictionary.
    table: Return a styled table trace with given headers and content.
    set_axis_limits: Set the axis limits for the given figure.
    get_axis_limits: Return the axis limits for the given figure.
    vector: Return a styled 2d or 3d vector trace from tail to head.
    scatter: Return a styled 2d or 3d scatter trace for given points and labels.
    line: Return a styled 2d line trace.
    intersection: Return the points where Ax = b intersects Dx <= e.
    equation: Return a styled 2d or 3d trace representing the given equation.
    order: Return an ordered list of points for drawing a 2d or 3d polygon.
    polygon: Return a styled 2d or 3d polygon trace defined by some points.
"""


def format(num: Union[int,float], precision: int = 3) -> str:
    """Return a properly formated string for a number at some precision."""
    return ('%.*f' % (precision, num)).rstrip('0').rstrip('.')


def linear_string(A: np.ndarray,
                  indices: List[int],
                  constant: float = None) -> str:
    """Return the string representation of a linear combination."""
    def sign(num: float): return {-1: ' - ', 0: ' + ', 1: ' + '}[np.sign(num)]
    s = ''
    if constant is not None:
        s += format(constant)
    for i in range(len(indices)):
        if i == 0:
            if constant is None:
                s += format(A[0]) + 'x<sub>' + str(indices[0]) + '</sub>'
            else:
                s += (sign(A[0]) + format(abs(A[0])) + 'x<sub>'
                      + str(indices[0]) + '</sub>')
        else:
            s += format(abs(A[i])) + 'x<sub>' + str(indices[i]) + '</sub>'
        if i is not len(indices)-1:
            s += sign(A[i+1])
    return s


def equation_string(A: np.ndarray, b: float, comp: str = ' ≤ ') -> str:
    """Return the string representation of an equation.

    The equation is assumed to be in standard form: Ax 'comp' b."""
    return linear_string(A, list(range(1, len(A) + 1))) + comp + format(b)


def label(dic: Dict[str, Union[float, list]]) -> str:
    """Return a styled string representation of the given dictionary."""
    entries = []
    for key in dic.keys():
        s = '<b>' + key + '</b>: '
        value = dic[key]
        if type(value) is float:
            s += format(value)
        if type(value) is list:
            s += '(%s)' % ', '.join(map(str, [format(i) for i in value]))
        entries.append(s)
    return '%s' % '<br>'.join(map(str, entries))


def table(header: List[str], content: List[str], style: str) -> plt.Table:
    """Return a styled table trace with given headers and content."""
    styles = ['canonical','dictionary']
    if style not in styles:
        raise ValueError("Invalid style. Currently supports " + styles)

    canon_args = dict(header=dict(values=header,
                                  height=30,
                                  font=dict(size=13),
                                  fill=dict(color='white'),
                                  line=dict(color='black', width=1)),
                      cells=dict(values=content,
                                 height=25,
                                 font=dict(size=13),
                                 fill=dict(color='white'),
                                 line=dict(color='black', width=1)))
    dict_args = dict(header=dict(values=header,
                                 height=25,
                                 font=dict(size=14),
                                 align=['left', 'right', 'left'],
                                 fill=dict(color='white'),
                                 line=dict(color='white', width=1)),
                     cells=dict(values=content,
                                height=25,
                                font=dict(size=14),
                                align=['left', 'right', 'left'],
                                fill=dict(color='white'),
                                line=dict(color='white', width=1)),
                     columnwidth=[0.3, 0.07, 0.63])
    return plt.Table({'canonical': canon_args, 'dictionary': dict_args}[style])


def set_axis_limits(fig: plt.Figure, x_list: List[np.ndarray]):
    """Set the axes limits of fig such that all points in x are visible.

    Given a set of nonnegative 2 or 3 dimensional points, set the axes
    limits such all points are visible within the plot window.
    """
    n = len(x_list[0])
    if n not in [2,3]:
        raise ValueError('x_list is a list of column vectors of length 2 or 3')
    pts = [list(x[:,0]) for x in x_list]
    limits = [max(i)*1.3 for i in list(zip(*pts))]
    if n == 2:
        x_lim, y_lim = limits
        pts = [np.array([[x_lim],[y_lim]])]
        fig.update_layout(xaxis=dict(range=[0, x_lim]),
                          yaxis=dict(range=[0, y_lim]))
        fig.update_layout(scene=dict(xaxis=dict(range=[0, x_lim]),
                                     yaxis=dict(range=[0, y_lim])))
    if n == 3:
        x_lim, y_lim, z_lim = limits
        pts = [np.array([[x_lim],[y_lim],[z_lim]])]
        fig.update_layout(scene=dict(xaxis=dict(range=[0, x_lim]),
                                     yaxis=dict(range=[0, y_lim]),
                                     zaxis=dict(range=[0, z_lim])))
    # Add an invisible point at the axes limits to prevent axes from rescaling
    fig.add_trace(scatter(pts, 'clear'))


def get_axis_limits(fig: plt.Figure,n: int) -> List[float]:
    """Return the axis limits for the given figure."""
    if n not in [2,3]:
        raise ValueError('Can only retrieve 2 or 3 axis limits')
    x_lim = fig.layout.scene.xaxis.range[1]
    y_lim = fig.layout.scene.yaxis.range[1]
    if n == 2:
        return x_lim, y_lim
    else:
        z_lim = fig.layout.scene.zaxis.range[1]
        return x_lim, y_lim, z_lim


def vector(tail: np.ndarray,
           head: np.ndarray) -> Union[plt.Scatter, plt.Scatter3d]:
    """Return a styled 2d or 3d vector trace from tail to head."""
    pts = list(zip(*[tail[:,0],head[:,0]]))
    if len(pts) == 2:
        x,y = pts
        z = None
    if len(pts) == 3:
        x,y,z = pts
    args = dict(x=x, y=y, mode='lines',
                line=dict(width=6, color='red'), opacity=1,
                hoverinfo='skip', showlegend=False, visible=False)
    if z is None:
        return plt.Scatter(args)
    else:
        args['z'] = z
        return plt.Scatter3d(args)


def scatter(x_list: List[np.ndarray],
            style: str,
            lbs: List[str] = None) -> plt.Scatter:
    """Return a styled 2d or 3d scatter trace for given points and labels."""
    styles = ['bfs', 'initial_sol', 'clear']
    if style not in styles:
        raise ValueError("Invalid style. Currently supports " + styles)

    pts = list(zip(*[list(x[:,0]) for x in x_list]))
    if len(pts) == 2:
        x,y = pts
        z = None
    if len(pts) == 3:
        x,y,z = pts

    bfs_args = dict(x=x, y=y, text=lbs, mode='markers',
                    marker=dict(size=20, color='gray', opacity=1e-7),
                    showlegend=False, hoverinfo='text',
                    hoverlabel=dict(bgcolor='#FAFAFA',
                                    bordercolor='#323232',
                                    font=dict(family='Arial',
                                              color='#323232')))
    init_args = dict(x=x, y=y, mode='markers',
                     marker=dict(size=5, color='red', opacity=1),
                     hoverinfo='skip', showlegend=False)
    clear_args = dict(x=x, y=y, mode='markers',
                      marker=dict(size=0, color='white', opacity=1e-7),
                      hoverinfo='skip', showlegend=False)

    args = {'bfs': bfs_args,
            'initial_sol': init_args,
            'clear': clear_args}[style]
    if z is None:
        return plt.Scatter(args)
    else:
        args['z'] = z
        return plt.Scatter3d(args)


def line(x_list: List[np.ndarray],
         style: str,
         lb: str = None,
         i=[0]) -> plt.Scatter:
    """Return a 2d line trace in the desired style."""
    styles = ['constraint', 'isoprofit']
    if style not in styles:
        raise ValueError("Invalid style. Currently supports " + styles)

    x,y = list(zip(*[list(x[:,0]) for x in x_list]))
    colors = ['#173D90', '#1469FE', '#65ADFF', '#474849', '#A90C0C', '#DC0000']
    if style == 'constraint':
        i[0] = i[0] + 1 if i[0] + 1 < 6 else 0

    con_args = dict(x=x, y=y, name=lb, mode='lines',
                    line=dict(color=colors[i[0]],
                              width=2,
                              dash='15,3,5,3'),
                    hoverinfo='skip', visible=True, showlegend=True)
    iso_args = dict(x=x, y=y, mode='lines',
                    line=dict(color='red', width=4, dash=None),
                    hoverinfo='skip', visible=False, showlegend=False)
    return plt.Scatter({'constraint': con_args, 'isoprofit': iso_args}[style])


def intersection(A: np.ndarray,
                 b: float,
                 D: np.ndarray,
                 e: float) -> List[np.ndarray]:
    """Return the points where Ax = b intersects Dx <= e."""
    n,m = len(A),len(D)
    if n not in [2,3]:
        raise ValueError('Only supports equations in 2 or 3 variables')
    lp = LP(np.vstack((D,A)),np.vstack((e,b)),np.ones((n,1)))
    pts = []
    for B in itertools.combinations(range(n+m),m+1):
        try:
            pts.append(lp.get_basic_feasible_sol(list(B)))
        except (InvalidBasis, InfeasibleBasicSolution):
            pass
    return [pt[0:n,:] for pt in pts]


def equation(fig: plt.Figure,
             A: np.ndarray,
             b: float,
             style: str,
             lb: str = None) -> Union[plt.Scatter, plt.Scatter3d]:
    """Return a styled 2d or 3d trace representing the given equation."""
    n = len(A)
    if n not in [2,3]:
        raise ValueError('Only supports equations in 2 or 3 variables')
    if all(A == np.zeros(n)):
        raise ValueError('A must have a nonzero component.')
    if n == 2:
        x_lim, y_lim = get_axis_limits(fig, n)
        # A[0]x + A[1]y = b
        if A[1] != 0:
            x = np.linspace(0,x_lim,2)
            y = (b - A[0]*x)/A[1]
            x_list = [np.array([[x[i]],[y[i]]]) for i in range(len(x))]
        else:
            x = b/A[0]
            x_list = [np.array([[x],[0]]),np.array([[x],[y_lim]])]
        return line(x_list,style,lb)
    if n == 3:
        x_lim, y_lim, z_lim = get_axis_limits(fig, n)
        # A[0]x + A[1]y + A[2]z = b
        x_list = []
        if A[2] != 0:
            for x in np.linspace(0,x_lim,2):
                for y in np.linspace(0,y_lim,2):
                    z = (b - A[0]*x - A[1]*y)/A[2]
                    x_list.append(np.array([[x],[y],[z]]))
        elif A[1] != 0:
            for x in np.linspace(0,x_lim,2):
                y = (b - A[0]*x)/A[1]
                for z in np.linspace(0,z_lim,2):
                    x_list.append(np.array([[x],[y],[z]]))
        else:
            x = b/A[0]
            for y in np.linspace(0,y_lim,2):
                for z in np.linspace(0,z_lim,2):
                    x_list.append(np.array([[x],[y],[z]]))
        return polygon(x_list,style,lb)


def order(x_list: List[np.ndarray]) -> List[List[float]]:
    """Return an ordered list of points for drawing a 2d or 3d polygon."""
    n,m = x_list[0].shape
    if not m == 1:
        raise ValueError('Points must be represented by column vectors')
    if n not in [2,3]:
        raise ValueError('Points must be 2 or 3 dimensional')

    pts = np.array([list(x[0:n,0]) for x in x_list])
    pts = np.unique(pts, axis=0)
    x_list = [np.array([pt]).transpose() for pt in pts]

    if len(pts) > 2:
        if n == 2:
            hull = ConvexHull(pts)
            pts = list(zip(pts[hull.vertices, 0],
                           pts[hull.vertices, 1]))
        if n == 3:
            b_1 = pts[1] - pts[0]
            b_2 = pts[2] - pts[0]
            b_3 = np.cross(b_1, b_2)
            T = np.linalg.inv(np.array([b_1, b_2, b_3]).transpose())
            x_list = [list(np.round(np.dot(T,x),7)[0:2,0]) for x in x_list]
            hull = ConvexHull(np.array(x_list))
            pts = list(zip(pts[hull.vertices, 0],
                           pts[hull.vertices, 1],
                           pts[hull.vertices, 2]))
            pts.append(pts[0])
    components = list(zip(*pts))
    components = [list(component) for component in components]
    return components


def polygon(x_list: List[np.ndarray],
            style: str,
            lb: str = None) -> plt.Scatter:
    """Return a styled 2d or 3d polygon trace defined by some points."""
    styles = ['region', 'constraint', 'isoprofit_in', 'isoprofit_out']
    if style not in styles:
        raise ValueError("Invalid style. Currently supports " + styles)
    if len(x_list) == 0:
        raise ValueError("The list of points was empty.")

    if len(x_list[0]) == 2:
        x,y = order(x_list)
        return plt.Scatter(x=x, y=y, mode='lines', fill='toself',
                           fillcolor='#1469FE', opacity=0.3,
                           line=dict(width=2, color='#00285F'),
                           showlegend=False, hoverinfo='skip')
    if len(x_list[0]) == 3:
        x,y,z = order(x_list)
        # When plotting a surface in Plotly, the surface is generated with
        # respect to a chosen axis. If the surface is orthogonal to this
        # axis, then the surface will not appear. This next step ensures
        # that each polygon surface will properly display.
        axis = 2  # default axis
        if len(x) > 2:
            # Get the normal vector of this polygon
            v1 = [x[1] - x[0], y[1] - y[0], z[1] - z[0]]
            v2 = [x[2] - x[0], y[2] - y[0], z[2] - z[0]]
            n = np.round(np.cross(v1,v2), 7)
            for ax in range(3):
                if not np.dot(n,[1 if i == ax else 0 for i in range(3)]) == 0:
                    axis = ax

        region_args = dict(x=x, y=y, z=z, surfaceaxis=axis,
                           surfacecolor='#1469FE', mode="lines",
                           line=dict(width=5, color='#173D90'),
                           opacity=0.2, hoverinfo='skip',
                           visible=True, showlegend=False)
        con_args = dict(x=x, y=y, z=z, name=lb, surfaceaxis=axis,
                        surfacecolor='gray', mode="none",
                        opacity=0.5, hoverinfo='skip',
                        visible='legendonly', showlegend=True)
        iso_in_args = dict(x=x, y=y, z=z, mode="lines+markers",
                           surfaceaxis=axis, surfacecolor='red',
                           marker=dict(size=5, color='red', opacity=1),
                           line=dict(width=5, color='red'),
                           opacity=1, hoverinfo='skip',
                           visible=False, showlegend=False)
        iso_out_args = dict(x=x, y=y, z=z, surfaceaxis=axis,
                            surfacecolor='gray', mode="none",
                            opacity=0.3, hoverinfo='skip',
                            visible=False, showlegend=False)
        return plt.Scatter3d({'region': region_args,
                              'constraint': con_args,
                              'isoprofit_in': iso_in_args,
                              'isoprofit_out': iso_out_args}[style])
